﻿using System;
using System.Configuration;
using System.Globalization;
using System.IdentityModel.Protocols.WSTrust;
using System.IdentityModel.Tokens;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Security;
using System.Security.Cryptography.X509Certificates;
using System.ServiceModel;
using System.ServiceModel.Configuration;
using System.ServiceModel.Security;
using Abr.AuskeyManager.KeyStore;
using USISampleCode.USIServiceReference;

namespace USISampleCode
{
    internal static class ServiceChannel
    {
        // Obviously these two variables should be in a config file. Included here so it's easier to see how this code works.
        private const string ClientCertificateKeystoreLocation = "keystore-usi.xml"; // @"C:\Developer\Userdata\keystore.xml";
        private const string Alias = "ABRD:27809366375_USIMachine"; // Old one: "ABRD:12300000059_TestDevice03";

        // This should be stored in encrypted form. See notes in GetPasswordString().
        private const string Password = "Password1!";

        private static ChannelFactory<IUSIService> _channelFactory;

        public static IUSIService OpenWithM2M()
        {
            var token = GetStsToken(60);
            var clientSection = (ClientSection) ConfigurationManager.GetSection("system.serviceModel/client");
            var endpointElement = clientSection.Endpoints.OfType<ChannelEndpointElement>()
                .First(endpoint => string.Equals("USIServiceReference.IUSIService", endpoint.Contract, StringComparison.OrdinalIgnoreCase));
            if (endpointElement == null)
            {
                throw new Exception("No endpoint matching service contract was found");
            }

            var channelFactory = new ChannelFactory<IUSIService>(endpointElement.Name);
            channelFactory.Open();
            return channelFactory.CreateChannelWithIssuedToken(token);
        }

        private static SecurityToken GetStsToken(int tokenLifeTimeMinutes)
        {
            // SecurityTokenServiceEndpoint vs S007SecurityTokenServiceEndpoint both support SAML1.1 & 2.0
            // see ATO's WSDL for difference between these 2
            var factory = new WSTrustChannelFactory("S007SecurityTokenServiceEndpoint");
            if (factory.Credentials != null)
            {
                factory.Credentials.ClientCertificate.Certificate = GetClientCertificateFromKeystore();

                // Instantiate and invoke the client to get the security token
                factory.Credentials.SupportInteractive = false;
            }

            var appliesTo = ConfigurationManager.AppSettings["appliesTo"];
            var rst = new RequestSecurityToken
            {
                Claims =
                {
                    new RequestClaim("http://vanguard.ebusiness.gov.au/2008/06/identity/claims/abn", false),
                    new RequestClaim("http://vanguard.ebusiness.gov.au/2008/06/identity/claims/credentialtype", false)
                },
                AppliesTo = new EndpointReference(appliesTo),
                Lifetime = new Lifetime(DateTime.UtcNow, DateTime.UtcNow.AddMinutes(tokenLifeTimeMinutes)),
                RequestType = RequestTypes.Issue,
                KeyType = KeyTypes.Symmetric,
                TokenType = "http://docs.oasis-open.org/wss/oasis-wss-saml-token-profile-1.1#SAMLV2.0",
                // TokenType = "http://docs.oasis-open.org/wss/oasis-wss-saml-token-profile-1.1#SAMLV1.1",
                // ActAs = This SecurityTokenElement need to be passed by the Host Service Providers.
            };

            // Instantiate and invoke the client to get the security token
            var client = (WSTrustChannel) factory.CreateChannel();
            var response = client.Issue(rst);
            return response;
        }

        private static X509Certificate2 GetClientCertificateFromKeystore()
        {
            var entryAssemblyName = Assembly.GetEntryAssembly()?.GetName();

            // Please replace [YourCompanyName] tag with the valid organisation Name
            AbrProperties.SetSoftwareInfo("[OrganisationName]", entryAssemblyName?.Name, entryAssemblyName?.Version.ToString(), DateTime.Now.ToString(CultureInfo.InvariantCulture));
            var keyStore = new AbrKeyStore(File.OpenRead(Path.Combine(AppDomain.CurrentDomain.BaseDirectory, ClientCertificateKeystoreLocation)));
            using (var pwd = GetPasswordString())
            {
                var abrCredential = keyStore.GetCredential(Alias);
                if (!abrCredential.IsReadyForRenewal())
                {
                    var clientCertificate = abrCredential.PrivateKey(pwd, X509KeyStorageFlags.MachineKeySet);
                    return clientCertificate;
                }
                else
                {
                    // throw new Exception("Renew certificate");
                    var clientCertificate = abrCredential.PrivateKey(pwd, X509KeyStorageFlags.MachineKeySet);
                    return clientCertificate;
                }
            }
        }

        private static SecureString GetPasswordString()
        {
            // NOTE: This code is for demonstration purposes only.
            //      Production code should obtain the password SecureString instance from an encrypted source,
            //      and it should never be held in a plain String object.
            //      Read MSDN remarks about why to avoid the password being stored in a plain String object: 
            //      http://msdn.microsoft.com/en-us/library/system.security.securestring%28v=vs.110%29.aspx
            var pwd = new SecureString();
            foreach (var c in Password)
            {
                pwd.AppendChar(c);
            }

            return pwd;
        }

        public static void Close()
        {
            _channelFactory?.Close();
            _channelFactory = null;
        }
    }
}